from collections import defaultdict
import json
import requests
import pandas as pd

from google.cloud import bigquery

from siametrics.deepmap.util import get_service_account_request_headers
from .param import Parameter


class DatasetDiscoveryService:
    service_url = 'https://dataset-discovery-service-whzo2dewrq-de.a.run.app'

    def __init__(self, g_cred, project_id):
        self.g_cred = g_cred
        self.project_id = project_id

    def make_request(self, data=None):
        data = json.dumps(data or {})

        url = DatasetDiscoveryService.service_url
        headers = get_service_account_request_headers(self.g_cred, url) if self.g_cred else None

        resp = requests.post(url, headers=headers, json=data)
        resp.raise_for_status()

        return resp

    def get_dataset(self, name):
        resp = self.make_request({'name': name})
        metadata = resp.json()
        return Dataset(name, project_id=self.project_id, metadata=metadata, g_cred=self.g_cred)

    def list_datasets(self):
        resp = self.make_request()
        return resp.json()['datasets']


class Dataset:
    def __init__(self, name, project_id, metadata, g_cred):
        self.name = name
        self.project_id = project_id
        self.__g_cred = g_cred

        self.__params = [Parameter.from_param_dict(param) for param in metadata['params']]

        assert metadata['query']['source'].lower() == 'bigquery'
        self.table = metadata['query']['params']['table']

        self.__schema = None
        self.__num_rows = None
        self.__num_bytes = None
        self.__description = None

    def __get_query(self, fields, **kwargs):
        query = {
            'table': self.table,
            'where': [],
        }

        for param in self.__params:
            name = param.name
            if name in kwargs:
                query = param.modify_query(query, kwargs[name])
            elif param.required:
                raise ValueError(f'Please provide required parameters: {param.name}')

        where_clause = ' AND '.join(query['where'])
        fields_string = ', '.join(fields)
        query_string = f"SELECT {fields_string} FROM {query['table']} WHERE {where_clause}"


        return query_string

    def fetch(self, fields=None, **kwargs):
        fields = fields or ['*']
        query = self.__get_query(fields, **kwargs)
        df = pd.read_gbq(query, project_id=self.project_id, credentials=self.__g_cred)
        return df

    @property
    def params(self):
        params = []
        for param in self.__params:
            record = {
                'name': param.name,
                'type': param.type_,
                'description': param.description,
            }
            params.append(record)
        return pd.DataFrame(params)

    @property
    def num_rows(self):
        if not self.__num_rows:
            self.__load_details()
        return self.__num_rows

    @property
    def num_bytes(self):
        if not self.__num_bytes:
            self.__load_details()
        return self.__num_bytes

    @property
    def description(self):
        if self.__description is None:
            self.__load_description()

        return self.__description

    def __load_schema(self, table):
        schema = []
        for s in table.schema:
            record = {
                'name': s.name,
                'type': s.field_type,
                'repeated': s.mode == 'REPEATED',
                'description': s.description,
            }
            schema.append(record)
        self.__schema = pd.DataFrame(schema).set_index('name')

    def __load_details(self):
        cred = self.__g_cred and self.__g_cred.with_scopes(['https://www.googleapis.com/auth/bigquery.readonly'])
        client = bigquery.Client(credentials=cred, project=self.project_id)

        dataset_id, table_id = self.table.split('.')
        dataset_ref = client.dataset(dataset_id, project=self.project_id)
        table_ref = dataset_ref.table(table_id)
        table = client.get_table(table_ref)

        self.__load_schema(table)
        self.__num_rows = table.num_rows
        self.__num_bytes = table.num_bytes

    def __get_stats(self, schema):
        select = []
        for name, s in schema.iterrows():
            if not s.repeated:
                select.append(f'MIN({s.name}) AS {s.name}__min')
                select.append(f'MAX({s.name}) AS {s.name}__max')
            select.append(f'COUNT({s.name}) AS {s.name}__count')
        query = 'SELECT ' + ', '.join(select) + f' FROM {self.table}'

        cred = self.__g_cred and self.__g_cred.with_scopes(['https://www.googleapis.com/auth/bigquery'])
        stats = pd.read_gbq(query, project_id=self.project_id, credentials=cred)

        aggs = defaultdict(dict)

        for key, value in stats.iloc[0].to_dict().items():
            name, agg = key.split('__')
            aggs[name][agg] = value

        stats = pd.DataFrame(aggs).T.sort_index(axis=1)
        return stats

    def __load_description(self):
        if self.__schema is None:
            self.__load_details()

        stats = self.__get_stats(self.__schema)
        description = pd.concat([self.__schema, stats], axis=1, sort=True)
        self.__description = description




import os
from shutil import copyfile
from google.cloud import iam_credentials_v1

default_key_path = os.path.join(os.path.expanduser('~'), '.deepmap', 'access-key.json')


def get_service_account_request_token(cred, url):
    cred = cred.with_scopes(['https://www.googleapis.com/auth/iam'])

    client = iam_credentials_v1.IAMCredentialsClient(credentials=cred)
    name = client.service_account_path('-', cred.service_account_email)
    token = client.generate_id_token(name, url).token
    return token


def get_service_account_request_headers(cred, url):
    token = get_service_account_request_token(cred, url)
    headers = {'Authorization': f'bearer {token}'}
    return headers


def install_key(file: str):
    dir_name = os.path.dirname(default_key_path)
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)

    copyfile(file, default_key_path)



class Parameter:
    def __init__(self, name, required, description, type_, template, arbitrary=False):
        self.name = name
        self.required = required
        self.description = description
        self.type_ = type_

        self.template = template
        self.arbitrary = arbitrary

    @classmethod
    def from_param_dict(cls, param):
        return cls(name=param['name'], required=param['required'],
                   description=param['description'], type_=param['type'], template=param['template'],
                   arbitrary=param.get('arbitrary', False))

    def modify_query(self, query, value):
        if 'where' in self.template:
            if self.type_.endswith('[]'):
                if self.arbitrary:
                    clause = self.template['where'].format(*value)
                else:
                    value = tuple(value) if len(value) != 1 else f'({repr(value[0])})'
                    clause = self.template['where'].format(value)
            else:
                clause = self.template['where'].format(value)

            query['where'].append(clause)
        else:
            raise NotImplementedError('Other template types are not implemented.')

        return query

    def __repr__(self):
        required = ' REQUIRED' if self.required else ''
        description = self.description[:50] + ('...' if len(self.description) > 50 else '')

        return f'dataset.Parameter(<{self.type_}>{self.name}{required}; {description})'



from setuptools import setup


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='siametrics-deepmap',
    version='0.0.2',
    packages=['siametrics', 'siametrics.deepmap', 'siametrics.deepmap.dataset'],
    url='',
    license='Apache 2.0',
    author='SSripilaipong',
    author_email='santhapon.s@siametrics.com',
    description='',
    install_requires=requirements,
)



import os
from typing import Union
from google.oauth2.service_account import Credentials

from siametrics.deepmap.dataset import DatasetDiscoveryService
from .util import default_key_path


class DeepmapClient:
    __default_project_id = 'siametrics-deepmap-dp'

    def __init__(self, key_path: Union[str, None] = None, project_id: Union[str, None] = None):
        self.__key_path = key_path or default_key_path
        self.project_id = project_id or DeepmapClient.__default_project_id

        if not os.path.exists(self.__key_path):
            self.__key_path = None

        self.__g_cred = None
        if self.__key_path:
            self.__g_cred = Credentials.from_service_account_file(self.__key_path)

        self.__discovery = DatasetDiscoveryService(self.__g_cred, self.project_id)

    def get_dataset(self, name: str):
        return self.__discovery.get_dataset(name)

    def list_datasets(self):
        return self.__discovery.list_datasets()
